#
# Copyright (c) 1991,1992 The Ohio State University.
# All rights reserved.
#
# Redistribution and use in source and binary forms are permitted
# provided that: (1) source distributions retain this entire copyright
# notice and comment, and (2) distributions including binaries display
# the following acknowledgement:  ``This product includes software
# developed by The Ohio State University and its contributors''
# in the documentation or other materials provided with the distribution
# and in all advertising materials mentioning features or use of this
# software. Neither the name of the University nor the names of its
# contributors may be used to endorse or promote products derived
# from this software without specific prior written permission.
# THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#

#
# show - show the contents of the tape or run database.
#

require 'global.defs';
require 'local.defs';
require $yagrip_perl;
require $backuplib_perl;
require $lib_misc_perl;
require $ctime_perl;

$options = "rvhd:";
$usage = 
"usage: To list entries in the tape or run databases:
  show [-r] [-v] [-d debuglevel] [-h] tape|run [ID]*
    
    -r                  reverse order of listing
    -v			verbose listing
    -d debuglevel	show debugging messages
    -h			display this message
    tape|run		show from either tape database or run database
    ID			tape or run ID numbers that you want to see (default
			is to display all)

To show runs that have been scheduled or started in the last N hours:
  show scheduled [N-hours]

To print a summary of the tape database:
  show [-d debuglevel] [-h] tape-summary\n";

    #
    # Make sure that we clean up after ourselves if someone wants to kill us.
    #
$SIG{'INT'} = 'handle_interrupts';
$SIG{'HUP'} = 'handle_interrupts';
$SIG{'TERM'} = 'handle_interrupts';
$SIG{'QUIT'} = 'handle_interrupts';

    #
    # Deal with command line arguments...
    #
die $usage if ! &getopt($options);
die $usage if $#ARGV < 0;
die $usage if defined($opt_h);

&read_config();

if ($ARGV[0] eq "tape-summary") {
    shift(@ARGV);
    die $usage if $#ARGV != -1;
    &tape_summary();
    exit(0);
}

if ($ARGV[0] eq "scheduled") {
    shift(@ARGV);
    if ($#ARGV == 0) {
	$hours = shift(@ARGV);
    } else {
	$hours = 24;
    }
    &scheduled_runs($hours);
    exit(0);
}

if ($ARGV[0] eq "tape" || $ARGV[0] eq "run") {
    die $usage if $#ARGV < 0;
    &list_tapes_or_runs(shift(@ARGV));
    exit(0);
}

die $usage;

#
# Report on the state of the tape library.  The report includes a 
# count of the free, used, expired and promised tapes. 
#

sub tape_summary {
    local(%used, %free, %promised, %expired, %bogus, $total);
    local(%types, %locations);
    local($tapeid, $tdb_value, %tdb_array, $flags, $type, $location);

    &open_db(*DB, $tape_db_file, $db_mode) ||
      die "error (show): can't open the database '$tape_db_file'\n";

    foreach $tapeid (&keys_db(*DB)) {
	&read_db(*DB, $tapeid, *tdb_value, *tdb_array);
    
	$flags = $tdb_array{"flags"};
	$type = $tdb_array{"type"};
	$location = $tdb_array{"location"} || "(default)";

	if (! defined($types{$type})) {
	    $types{$type} = 1;
	}
	if (! defined($locations{$location})) {
	    $locations{$location} = 1;
	}

	if ($flags eq "used") {
	    $used{$location,$type}++;
	} elsif ($flags eq "free") {
	    $free{$location,$type}++;
	} elsif ($flags eq "promised") {
	    $promised{$location,$type}++;
	} elsif ($flags eq "expired") {
	    $expired{$location,$type}++;
	} else {
	    print "warning (show): Gack! tape $tapeid has a flag of '$flags'!\n";
	    $bogus{$location,$type}++;
	}
    }

    &close_db(*DB, $tape_db_file);

    foreach $location (sort keys %locations) {
	print "location $location\n";
	foreach $type (sort keys %types) {
	    printf "  %s\t%d free, %d promised, %d used, %d expired, %d bogus, %d total\n",
	    $type,
	    $free{$location,$type},
	    $promised{$location,$type},
	    $used{$location,$type},
	    $expired{$location,$type},
	    $bogus{$location,$type},
	    $free{$location,$type}+$promised{$location,$type}+
	      $used{$location,$type}+$expired{$location,$type}+
		$bogus{$location,$type};
	}
    }	    
}

sub list_tapes_or_runs {
    local($database) = @_;
    local($db_file, @keywords, $id, $db_value, %db_array);

    if ($database eq "tape") {
	$db_file = $tape_db_file;
    } elsif ($database eq "run") {
	$db_file = $run_db_file;
    } else {
	die "error (show): internal error, &list_tapes_or_runs got database '$database'\n";
    }

    #
    # Do the work...open the database, for each keyword given (all
    # if none were specified) call print_tape_entry to show the entry.
    #
    &open_db(*DB, $db_file, $db_mode) ||
      die "error (show): can't open the database $db_file on $hostname in $0\n";

    if ($#ARGV >= 0) {
	@keywords = @ARGV;
    } else {
	@keywords = &keys_db(*DB);
    }

    if ($database eq "tape") {
	@keywords = sort increasing_tapeid @keywords;
    } else {
	@keywords = sort increasing @keywords;
    }

    if (defined($opt_r)) {
	@keywords = reverse @keywords;
    }

    foreach $id (@keywords) {
        if (&defined_db(*DB, $id)) {
	    &read_db(*DB, $id, *db_value, *db_array);

	    if ($database eq "tape") {
		&print_tape_entry($id, %db_array);
	    } elsif ($database eq "run") {
		&print_run_entry($id, %db_array);
	    } else {
		warn "internal error - bad database type $database\n";
	    }
	} else {
	    warn "warning (show): there's no entry for $database $id in the database.\n";
	}
    }

    &close_db(*DB, $db_file);
}

sub scheduled_runs {
    local($hours) = @_;
    local($db_file, @keywords, $id, $db_value, %db_array);

    #
    # Do the work...open the database, for each keyword given (all
    # if none were specified) check the date and call print_run_entry 
    # to show the entry. 
    #
    &open_db(*DB, $run_db_file, $db_mode) ||
      die "error (show): can't open the database $run_db_file on $hostname\n";

    @keywords = sort increasing &keys_db(*DB);

    foreach $id (@keywords) {
        if (&defined_db(*DB, $id)) {
	    &read_db(*DB, $id, *db_value, *db_array);

	    if ((((time - $db_array{"start_date"}) / (60 * 60)) < $hours) &&
		($db_array{"flags"} eq "scheduled" || 
		 $db_array{"flags"} eq "started")) {
		&print_run_entry($id, %db_array);
	    }
	} else {
	    warn "warning (show): there's no entry for $database $id in the database.\n";
	}
    }

    &close_db(*DB, $run_db_file);
}

